﻿Imports System.Windows.Controls
Imports DataFields
Imports Microsoft.Win32
Imports System.Configuration

Public Class DataEntryWindow
    Private InitialWindowWidth As Double
    Private InitialGridWidth As Double
    Private InitialValueWidth As Double
    Private InitialDirectory As String = ConfigurationSettings.AppSettings("TemplateFolder")
    Private InitialOutputDirectory As String = ConfigurationSettings.AppSettings("OutputFolder")
    Private DisplayFields As WorkflowFields

    Public Property Fields() As Dictionary(Of String, String)
    Public Property Document() As String

 
    ''' <summary>
    ''' Initial Actions on Data Entry Window Loaded
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub DataEntryWindow_Loaded(ByVal sender As Object, ByVal e As System.Windows.RoutedEventArgs) Handles Me.Loaded
        InitialWindowWidth = Me.Width
        InitialGridWidth = DataGridVariables.Width
        InitialValueWidth = DataGridVariables.Columns(1).Width.Value
        If Fields Is Nothing Then Fields = New Dictionary(Of String, String)
        ' If a document path is passed in, use it

        If Not String.IsNullOrEmpty(Me.Document) Then
            Fields = GetFieldsForDocument(Me.Document)
            ' Allow user to enter the data from the initial document

            ' Add Input document
            If Not Fields.ContainsKey("InputDocument") Then Fields.Add("InputDocument", Document)

            ' Add output document 
            If Not Fields.ContainsKey("OutputDocument") Then Fields.Add("OutputDocument", String.Empty)

            ' Add Target Activity if not present
            If Not Fields.ContainsKey("TargetActivity") Then Fields.Add("TargetActivity", String.Empty)

            FillGrid(Fields, DataGridVariables)
            Me.btnEnterData.IsEnabled = True
        End If
    End Sub
    ''' <summary>
    ''' Return False if no document was selected
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub DataEntryWindow_Closed(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Closed
        If String.IsNullOrEmpty(Document) Then Me.DialogResult = False
    End Sub
   
    ''' <summary>
    ''' Put fields into the workflow
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub btnEnterData_Click(ByVal sender As Object, ByVal e As System.Windows.RoutedEventArgs) Handles btnEnterData.Click
        Dim CurrentCursor As Cursor = Me.Cursor
        Me.Cursor = Cursors.Wait
        If Me.DataGridVariables.Items.Count = 0 Then
            MessageBox.Show("No data")
            Me.Cursor = CurrentCursor
        End If
        '
        '   Require an output document
        Dim OutputDocument As String = String.Empty
        Dim OutputField As WorkFlowField = Nothing
        If DisplayFields IsNot Nothing AndAlso DisplayFields.ContainsKey("OutputDocument") Then
            OutputField = DisplayFields.Item("OutputDocument")
        End If
        If OutputField IsNot Nothing Then
            OutputDocument = OutputField.Value
        End If

        If OutputDocument = String.Empty Then
            MessageBox.Show("No output document defined")
            btnEnterData.IsEnabled = True
            Me.Cursor = CurrentCursor
            Exit Sub
        End If

        'Copy internal DisplayFields values to Fields property values
        For Each Field As WorkFlowField In DisplayFields
            Fields(Field.Key) = Field.Value
        Next

        Me.Cursor = CurrentCursor
        Me.DialogResult = True
        Me.Close()
    End Sub
    ''' <summary>
    ''' Allow user to select a document as a prototype
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>  
    Private Sub btnSelectDocument_Click(ByVal sender As Object, ByVal e As System.Windows.RoutedEventArgs) Handles btnSelectDocument.Click
        btnEnterData.IsEnabled = True
        Dim OpenFile As New OpenFileDialog
        With OpenFile
            .DefaultExt = "*.*"
            .Filter = "Word (*.docx, *.docm)|*.docx; *.docm|Excel (*.xlsx, *.xlsm)|*.xlsx;*.xlsm"
            If String.IsNullOrEmpty(InitialDirectory) Then InitialDirectory = Environment.SpecialFolder.MyDocuments.ToString()
            .InitialDirectory = InitialDirectory
            .CheckFileExists = True
            .CheckPathExists = True

            ' Dialog box is set to check for file exists and path exists...
            If .ShowDialog Then
                Document = .FileName
            Else
                Exit Sub
            End If
        End With

        If String.IsNullOrEmpty(Document) Then
            MessageBox.Show("Can't execute workflow without a document", "Data Entry", MessageBoxButton.OK, MessageBoxImage.Warning)
            Exit Sub
        End If

        ' Get the fields defined in the input document
        Fields = GetFieldsForDocument(Document)

        ' Add Input document
        If Not Fields.ContainsKey("InputDocument") Then Fields.Add("InputDocument", Document)

        ' Add output document directory
        If String.IsNullOrEmpty(InitialOutputDirectory) Then InitialOutputDirectory = String.Empty
        If Not Fields.ContainsKey("OutputDocument") Then Fields.Add("OutputDocument", InitialOutputDirectory)

        ' Add Target Activity if not present
        If Not Fields.ContainsKey("TargetActivity") Then Fields.Add("TargetActivity", String.Empty)

        FillGrid(Fields, DataGridVariables)
        btnSelectDocument.IsEnabled = False
    End Sub
    ''' <summary>
    ''' Get the field names for the document 
    ''' </summary>
    ''' <remarks>
    ''' Based on the type of document, the public fields are found by use of the appropriate Office document type
    ''' </remarks>
    Private Function GetFieldsForDocument(ByVal inputDocument As String) As Dictionary(Of String, String)

        If Not String.IsNullOrEmpty(inputDocument) Then
            Document = inputDocument
        End If

        Me.btnEnterData.IsEnabled = True
        Dim FileType As String = System.IO.Path.GetExtension(Document).ToLowerInvariant
        Dim Message As String = String.Format("Document={0}, FileType={1}", Document, FileType)

        ' Handle Word, Excel document types
        Select Case FileType
            Case ".docx", "*.docm"
                Return GetFieldsForWord(Document, Me.DataGridVariables)
            Case ".xlsx", ".xlsm"
                Return GetFieldsForExcel(Document, Me.DataGridVariables)
            Case Else
                Throw New ArgumentOutOfRangeException("Unknown file type")
        End Select
    End Function
    ''' <summary>
    ''' Get the fields (Bookmarks) for a word document 
    ''' </summary>
    ''' <param name="filePath">Path to document</param>
    ''' <param name="grid">DataGridView to populate</param>
    ''' <remarks></remarks>
    Private Function GetFieldsForWord(ByVal filePath As String, ByRef grid As DataGrid) As Dictionary(Of String, String)
        Dim WordSource As New OfficeInterfaces.WordInterface
        Dim WordFields As Dictionary(Of String, String) = WordSource.GetFieldNames(filePath)
        WordSource.Dispose()
        Return WordFields
    End Function
    ''' <summary>
    ''' Get the fields corresponding to the Excel named ranges
    ''' </summary>
    ''' <param name="filePath"></param>
    ''' <param name="grid"></param>
    ''' <remarks></remarks>
    Private function GetFieldsForExcel(ByVal filePath As String, ByRef grid As DataGrid) As Dictionary(Of String, String)
        Dim ExcelSource As New OfficeInterfaces.ExcelInterface
        Dim ExcelFields As Dictionary(Of String, String) = ExcelSource.GetFieldNames(filePath)
        ExcelSource.Dispose()
        Return ExcelFields
    End Function
    ''' <summary>
    ''' Fill the provided DataGrid from the provided Dictionary
    ''' </summary>
    ''' <param name="Fields"></param>
    ''' <param name="grid"></param>
    ''' <remarks>
    ''' </remarks>
    Private Sub FillGrid(ByVal Fields As Dictionary(Of String, String), ByRef grid As DataGrid)
        grid.Items.Clear()
        DisplayFields = New WorkflowFields(Fields)
        grid.ItemsSource = DisplayFields
    End Sub
    ''' <summary>
    ''' Re-size grid width when window is re-sized
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub DataEntryWindow_SizeChanged(ByVal sender As Object, ByVal e As System.Windows.SizeChangedEventArgs) Handles Me.SizeChanged
        If Me.WindowState = System.Windows.WindowState.Minimized Then Exit Sub
        Select Case True
            Case Me.Width > InitialWindowWidth
                DataGridVariables.Width = InitialGridWidth + Me.Width - InitialWindowWidth
            Case Me.Width < InitialWindowWidth
                Me.Width = e.PreviousSize.Width
            Case Me.Width = InitialWindowWidth
                DataGridVariables.Width = InitialGridWidth
        End Select
    End Sub
    ''' <summary>
    ''' Enable single-click editing
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub DataGridVariables_SelectedCellsChanged(ByVal sender As Object, ByVal e As System.Windows.Controls.SelectedCellsChangedEventArgs) Handles DataGridVariables.SelectedCellsChanged
        DataGridVariables.BeginEdit()
    End Sub
End Class
