﻿Imports System.IO
Imports System.IO.Packaging
Imports System.Linq
Imports DocumentFormat.OpenXml.Packaging
Imports DocumentFormat.OpenXml.Spreadsheet
Imports DocumentFormat.OpenXml
'Imports <xmlns:x='http://schemas.openxmlformats.org/spreadsheetml/2006/main'>
Imports System.Text.RegularExpressions
Imports System.Globalization

''' <summary>
''' Provides the ability to fill in an Excel workbook
''' </summary>
''' <remarks></remarks>
Public Class ExcelInterface
    Inherits OfficeInterface
    Dim _SpreadSheetDocument As DocumentFormat.OpenXml.Packaging.SpreadsheetDocument

    ''' <summary>
    ''' Default Constructor
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        MyBase.new()
    End Sub
    ''' <summary>
    ''' Document only constructor
    ''' </summary>
    ''' <param name="document"></param>
    ''' <remarks></remarks>
    Public Sub New(ByVal document As String)
        Me.New(document, Nothing)
    End Sub
    ''' <summary>
    ''' Full Constructor
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="fields"></param>
    ''' <remarks></remarks>
    Public Sub New(ByVal document As String, ByVal fields As Dictionary(Of String, String))
        MyBase.new()
        Me.Document = document
        Me.Fields = fields
        ' Mark document as not to be autosaved - need to save a new version
        Dim Settings As New DocumentFormat.OpenXml.Packaging.OpenSettings
        Settings.AutoSave = False
        _SpreadSheetDocument = CreateSpreadSheetDocument(document, True, Settings)
    End Sub
    ''' <summary>
    ''' Spreadsheet
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property SpreadSheetDocument As SpreadsheetDocument
        Get
            Return _SpreadSheetDocument
        End Get
        Set(ByVal value As SpreadsheetDocument)
            _SpreadSheetDocument = value
        End Set
    End Property

    Public Overrides Sub Close()
    End Sub
    ''' <summary>
    ''' Fill in Spreadsheet document ranges whose names match the fields - path to document and fields provided
    ''' </summary>
    ''' <param name="document">string, path to document</param>
    ''' <param name="fields"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function FillInDocument(ByVal document As String, ByVal fields As System.Collections.Generic.Dictionary(Of String, String)) As Boolean
        Me.Document = document
        Me.Fields = fields
        If fields Is Nothing Then Throw New ArgumentNullException("Fields")
        If Not System.IO.File.Exists(document) Then Throw New ArgumentNullException("Document")
        ' Mark document as not to be autosaved - need to save a new version
        Dim Settings As New DocumentFormat.OpenXml.Packaging.OpenSettings
        Settings.AutoSave = False
        _SpreadSheetDocument = Me.CreateSpreadSheetDocument(document, True, Settings)
        Return FillInDocument()
    End Function
    ''' <summary>
    ''' Fill in Spreadsheet document ranges whose names match the fields - SpreadSheetDocument and fields provided
    ''' </summary>
    ''' <param name="spreadSheetDocument"></param>
    ''' <param name="fields"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overloads Function FillInDocument(ByVal spreadSheetDocument As SpreadsheetDocument, ByVal fields As Dictionary(Of String, String)) As Boolean
        _SpreadSheetDocument = spreadSheetDocument
        Me.Fields = fields
        Return FillInDocument()
    End Function
    ''' <summary>
    ''' Fill in the Spreadsheet document ranges whose names match the fields
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function FillInDocument() As Boolean
        Dim Status As Boolean = False
        If SpreadSheetDocument Is Nothing Then Throw New ArgumentOutOfRangeException("SpreadSheetDocument")

        ' Find named ranges matching Fields
        Dim Workbook As Workbook = SpreadSheetDocument.WorkbookPart.Workbook

        Dim ActiveWorksheetPart As WorksheetPart = Nothing

        ' Check each Named Range for a matching Field
        For Each NamedRange As DefinedName In Workbook.DefinedNames
            Dim Name As String = NamedRange.Name
            Dim NewValue As String = String.Empty
            ' Defined range (e.g. "Sheet1!$B$4")
            Dim RangeLocation As String = NamedRange.Text
            Dim RangePartNames() As String = Split(RangeLocation, "!", 2, CompareMethod.Text)
            If RangePartNames Is Nothing Then Throw New InvalidDataException(String.Format("No location found for Range {0}", Name))
            Dim SheetName As String = String.Empty
            Dim CellName As String = String.Empty
            Dim UriString As String = String.Empty
            Select Case RangePartNames.Length
                Case 2
                    SheetName = RangePartNames(0)
                    CellName = RangePartNames(1)
                Case 1
                    CellName = RangePartNames(0)
                Case Else
                    Throw New InvalidDataException(String.Format("No location found for Range {0}", Name))
            End Select

            If String.IsNullOrEmpty(SheetName) Then Throw New ArgumentException("Null sheet name")

            ' If this range name doesn't have a matching input field, check for special cases
            If Me.Fields.ContainsKey(Name) Then
                NewValue = Fields(Name)
            Else
                Select Case Name
                    Case "FullName"
                        NewValue = GetFullName(Fields)
                    Case "FullAddress"
                        NewValue = GetFullAddress(Fields)
                    Case Else
                        Continue For
                End Select
            End If

            ' Find the worksheet which contains the named range
            For Each Part As WorksheetPart In SpreadSheetDocument.WorkbookPart.WorksheetParts
                UriString = Part.Uri.ToString
                Dim i As Integer = UriString.LastIndexOf("/")
                Dim j As Integer = UriString.LastIndexOf(".")
                'Parse the WorksheetPart's Uri to match it with the sheet name
                If UriString.Substring(i + 1, j - i - 1) = SheetName.ToLower Then
                    ActiveWorksheetPart = Part
                    Exit For
                End If
            Next Part
            If ActiveWorksheetPart Is Nothing Then Throw New ArgumentNullException("Worksheet is null")

            Dim TargetCell As Cell = InsertCellInWorksheet(GetColumnName(CellName), GetRowIndex(CellName), ActiveWorksheetPart)
            ' Check data type in cell
            If TargetCell.DataType Is Nothing Then  'This is ordinary data directly in the cell
                'Test for a string - not a numeric value, not a date
                If Not IsNumeric(NewValue) AndAlso Not IsDate(NewValue) Then
                    Dim index As Integer = InsertSharedStringItem(NewValue, SpreadSheetDocument)
                    TargetCell.CellValue = New CellValue(index.ToString)
                    TargetCell.DataType = New EnumValue(Of CellValues)(CellValues.SharedString)
                    Status = True
                    Continue For
                End If
                TargetCell.CellValue = New CellValue(NewValue)
                Status = True

                Dim worksheet As Worksheet = ActiveWorksheetPart.Worksheet
                Dim sheetData As SheetData = worksheet.GetFirstChild(Of SheetData)()
                ' Look for an associated implicit cell format
                If TargetCell.StyleIndex IsNot Nothing AndAlso TargetCell.StyleIndex.HasValue Then
                    For Each CellAttribute As OpenXmlAttribute In SpreadSheetDocument.WorkbookPart.WorkbookStylesPart.Stylesheet.CellFormats(TargetCell.StyleIndex.Value).GetAttributes
                        If CellAttribute.LocalName = "numFmtId" Then
                            Dim FormatIndex As Integer = Integer.Parse(CellAttribute.Value)
                            ' Handle special cases of implicit formats
                            Select Case FormatIndex
                                Case 14 To 22    'Date/time formats
                                    If IsDate(NewValue) Then
                                        Dim ExcelDate As String = Me.ConvertDateTimeToExcelDate(DateTime.Parse(NewValue, CultureInfo.CurrentCulture))
                                        TargetCell.CellValue.Text = ExcelDate
                                    End If
                            End Select
                            Exit For
                        End If
                    Next
                End If
            ElseIf TargetCell.DataType.HasValue Then
                Select Case TargetCell.DataType.Value
                    Case CellValues.SharedString    ' Handle case of a shared string data type
                        ' Insert the text into the SharedStringTablePart.
                        Dim index As Integer = InsertSharedStringItem(NewValue, SpreadSheetDocument)
                        TargetCell.CellValue.Text = index.ToString
                        Status = True
                End Select
            End If
        Next NamedRange

        ' Handle the case where no matching range was found
        If Status = False Then Return Status
        ' Save the document if a SaveAs location was provided

        ActiveWorksheetPart.Worksheet.Save() 'Update the contents regardless

        If String.IsNullOrEmpty(Me.SaveAs) Then Return Status
        SaveDocument()
        Return Status
    End Function
    ''' <summary>
    ''' Get the location of a named range in specified file
    ''' </summary>
    ''' <param name="rangeName"></param>
    ''' <param name="documentPath"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetRangeLocation(ByVal rangeName As String, ByVal documentPath As String) As String
        If String.IsNullOrEmpty(documentPath) Then Throw New ArgumentNullException("document")
        If Not System.IO.File.Exists(documentPath) Then Throw New FileNotFoundException("document")
        _SpreadSheetDocument = CreateSpreadSheetDocument(documentPath, False)
        Return GetRangeLocation(rangeName)
    End Function
    ''' <summary>
    ''' Get the location of a named range in specified SpreadSheetDocument
    ''' </summary>
    ''' <param name="rangeName"></param>
    ''' <param name="spreadsheetDocument"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetRangeLocation(ByVal rangeName As String, ByVal spreadsheetDocument As SpreadsheetDocument) As String
        Me._SpreadSheetDocument = spreadsheetDocument
        Return GetRangeLocation(rangeName)
    End Function
    ''' <summary>
    ''' Get the location of a named range in existing SpreadSheetDocument
    ''' </summary>
    ''' <param name="rangeName"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetRangeLocation(ByVal rangeName As String) As String
        ' Find named range matching input name
        Dim Workbook As Workbook = SpreadSheetDocument.WorkbookPart.Workbook
        Dim ActiveWorksheetPart As WorksheetPart = Nothing
        Dim RangeLocation As String = String.Empty
        For Each NamedRange As DefinedName In Workbook.DefinedNames
            Dim Name As String = NamedRange.Name
            If Name = rangeName Then
                RangeLocation = NamedRange.Text
                Exit For
            End If
        Next
        Return RangeLocation
    End Function
    ''' <summary>
    ''' Get a Dictionary of named ranges in the target SpreadSheet document file 
    ''' </summary>
    ''' <param name="documentPath">Path to target document</param>
    ''' <returns>Dictionary (name, value)</returns>
    ''' <remarks></remarks>
    Public Overrides Function GetFieldNames(ByVal documentPath As String) As Dictionary(Of String, String)
        If String.IsNullOrEmpty(documentPath) Then Throw New ArgumentNullException("document")
        If Not System.IO.File.Exists(documentPath) Then Throw New FileNotFoundException("document")
        Me.Document = Document
        _SpreadSheetDocument = CreateSpreadSheetDocument(documentPath, False)

        Return GetFieldNames(SpreadSheetDocument)
    End Function
    ''' <summary>
    ''' Get a Dictionary of named ranges, values in the existing SpreadSheetDocument
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overloads Function GetFieldNames() As Dictionary(Of String, String)
        If SpreadSheetDocument Is Nothing Then Throw New ArgumentNullException("SpreadSheetDocument")
        Return GetFieldNames(SpreadSheetDocument)
    End Function
    ''' <summary>
    ''' Get a Dictionary of named ranges, values in the target SpreadsheetDocument
    ''' </summary>
    ''' <param name="spreadSheetDocument"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overloads Function GetFieldNames(ByVal spreadsheetDocument As SpreadsheetDocument) As Dictionary(Of String, String)
        Dim Ranges As New Dictionary(Of String, String)
        Dim SheetName As String = String.Empty
        Dim WorksheetPart As WorksheetPart = Nothing
        ' Extract named ranges
        Dim Workbook As Workbook = spreadsheetDocument.WorkbookPart.Workbook

        For Each NamedRange As DefinedName In Workbook.DefinedNames
            Dim Name As String = NamedRange.Name
            Dim Value As String = String.Empty
            ' Defined range (e.g. "Sheet1!$B$4")
            Dim RangeLocation As String = NamedRange.Text
            Dim RangePartNames() As String = Split(RangeLocation, "!", 2, CompareMethod.Text)
            If RangePartNames Is Nothing Then Throw New InvalidDataException(String.Format("No location found for Range {0}", Name))
            Dim CellName As String = String.Empty
            Select Case RangePartNames.Length
                Case 2
                    SheetName = RangePartNames(0)
                    CellName = RangePartNames(1)
                Case 1
                    CellName = RangePartNames(0)
                Case Else
                    Throw New InvalidDataException(String.Format("No location found for Range {0}", Name))
            End Select
            If String.IsNullOrEmpty(SheetName) Then
                ' This may not occur
                Throw New ArgumentException("Null sheet name")
            Else
                For Each Part As WorksheetPart In spreadsheetDocument.WorkbookPart.WorksheetParts
                    Dim UriString As String = Part.Uri.ToString
                    Dim i As Integer = UriString.LastIndexOf("/")
                    Dim j As Integer = UriString.LastIndexOf(".")

                    If UriString.Substring(i + 1, j - i - 1) = SheetName.ToLower Then
                        WorksheetPart = Part
                        Exit For
                    End If
                Next Part
            End If
            If WorksheetPart Is Nothing Then Throw New ArgumentNullException("Worksheet is null")
            Dim TargetCell As Cell = InsertCellInWorksheet(GetColumnName(CellName), GetRowIndex(CellName), WorksheetPart)
            Dim CellFormat As String = String.Empty
            ' Check data type in cell
            If TargetCell.DataType Is Nothing Then  'This is ordinary data directly in the cell
                If TargetCell.CellValue Is Nothing OrElse String.IsNullOrEmpty(TargetCell.CellValue.InnerText) Then
                    Value = String.Empty
                Else
                    Value = TargetCell.CellValue.InnerText
                    ' Look for an associated implicit cell format
                    If TargetCell.StyleIndex IsNot Nothing AndAlso TargetCell.StyleIndex.HasValue Then
                        For Each CellAttribute As OpenXmlAttribute In spreadsheetDocument.WorkbookPart.WorkbookStylesPart.Stylesheet.CellFormats(TargetCell.StyleIndex.Value).GetAttributes
                            If CellAttribute.LocalName = "numFmtId" Then
                                Dim FormatIndex As Integer = Integer.Parse(CellAttribute.Value)
                                CellFormat = Me.GetImplicitCellFormat(FormatIndex)
                                ' Handle special cases of implicit formats, modified for correct .NET display 
                                Select Case FormatIndex
                                    Case 14 To 22    'Date/time formats
                                        Dim DateValue As Date = Me.ConvertExcelDateToDateTime(Value)
                                        If Me.DateFormats.ContainsKey(CellFormat) Then
                                            CellFormat = DateFormats(CellFormat)
                                        End If
                                        Value = DateValue.ToString(CellFormat)
                                End Select
                                Exit For
                            End If
                        Next
                    End If
                End If
            ElseIf TargetCell.DataType.HasValue Then
                Select Case TargetCell.DataType.Value
                    Case CellValues.SharedString
                        Dim sharedStringId As Integer = Integer.Parse(TargetCell.CellValue.Text)
                        Dim shareStringTablePart As SharedStringTablePart = spreadsheetDocument.WorkbookPart.SharedStringTablePart
                        Dim item As SharedStringItem = shareStringTablePart.SharedStringTable.Elements(Of SharedStringItem)().ElementAt(sharedStringId)
                        Value = item.InnerText
                End Select
            End If
            If Not String.IsNullOrEmpty(Name) AndAlso Not Ranges.ContainsKey(Name) Then
                Ranges.Add(Name, Value)
            End If
        Next NamedRange
        Return Ranges
    End Function

    ' Row where the data in B3 is a date. s="1" may look like
    '<row r="3" spans="1:2">
    '   <c r="A3" t="s">
    '       <v>1</v> 
    '       </c>
    '   <c r="B3" s="1">
    '       <v>40372</v> 
    '       </c>
    '</row>
    '<x:xf numFmtId="14" fontId="0" fillId="0" borderId="0" xfId="0" applyNumberFormat="1" xmlns:x="http://schemas.openxmlformats.org/spreadsheetml/2006/main" />

    ''' <summary>
    ''' Create a Spreadsheet document in a memory stream from a file
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="isEditable"></param>
    ''' <param name="settings"></param>
    ''' <remarks>
    ''' Side effects - creates a Package as well as a memory stream representation 
    ''' </remarks>
    Private Function CreateSpreadSheetDocument(ByVal document As String, ByVal isEditable As Boolean, Optional ByVal settings As OpenSettings = Nothing) As SpreadsheetDocument
        Me.CloseSpreadSheetDocument()
        Me.CloseDocumentPackage()
        Me.Document = document
        LoadDocumentToMemoryStream(document)
        Me.DocumentPackage = Package.Open(DocumentStream, FileMode.Open, FileAccess.ReadWrite)
        If settings Is Nothing Then
            Return SpreadSheetDocument.Open(DocumentPackage)
        Else
            Return SpreadSheetDocument.Open(DocumentPackage, settings)
        End If
    End Function

    ''' <summary>
    ''' Delete text from a cell in a document from a file
    ''' </summary>
    ''' <param name="docName"></param>
    ''' <param name="sheetName"></param>
    ''' <param name="colName"></param>
    ''' <param name="rowIndex"></param>
    ''' <remarks>
    ''' Given a document, a worksheet name, a column name, and a one-based row index,
    ''' deletes the text from the cell at the specified column and row on the specified sheet.
    ''' From: Open XML SDK 2.0
    ''' </remarks>
    Public Sub DeleteTextFromCell(ByVal docName As String, ByVal sheetName As String, ByVal colName As String, ByVal rowIndex As UInteger)
        ' Open the document for editing.
        Dim document As SpreadsheetDocument = SpreadSheetDocument.Open(docName, True)
        DeleteTextFromCell(document, sheetName, colName, rowIndex)
    End Sub
    ''' <summary>
    ''' Insert or locate a cell in a WorksheetPart
    ''' </summary>
    ''' <param name="columnName"></param>
    ''' <param name="rowIndex"></param>
    ''' <param name="worksheetPart"></param>
    ''' <returns>Cell</returns>
    ''' <remarks>
    ''' Given a column name, a row index, and a WorksheetPart, inserts a cell into the worksheet. 
    ''' If the cell already exists, return it. 
    ''' From: Open XML SDK 2.0
    ''' </remarks>
    Private Function InsertCellInWorksheet(ByVal columnName As String, ByVal rowIndex As UInteger, ByVal worksheetPart As WorksheetPart) As Cell
        Dim worksheet As Worksheet = worksheetPart.Worksheet
        Dim sheetData As SheetData = worksheet.GetFirstChild(Of SheetData)()
        Dim cellReference As String = (columnName + rowIndex.ToString())

        ' If the worksheet does not contain a row with the specified row index, insert one.
        Dim row As Row
        If (sheetData.Elements(Of Row).Where(Function(r) r.RowIndex.Value = rowIndex).Count() <> 0) Then
            row = sheetData.Elements(Of Row).Where(Function(r) r.RowIndex.Value = rowIndex).First()
        Else
            row = New Row()
            row.RowIndex = rowIndex
            sheetData.Append(row)
        End If

        ' If there is not a cell with the specified column name, insert one.  
        If (row.Elements(Of Cell).Where(Function(c) c.CellReference.Value = columnName + rowIndex.ToString()).Count() > 0) Then
            Return row.Elements(Of Cell).Where(Function(c) c.CellReference.Value = cellReference).First()
        Else
            ' Cells must be in sequential order according to CellReference. Determine where to insert the new cell.
            Dim refCell As Cell = Nothing
            For Each cell As Cell In row.Elements(Of Cell)()
                If (String.Compare(cell.CellReference.Value, cellReference, True) > 0) Then
                    refCell = cell
                    Exit For
                End If
            Next

            Dim newCell As Cell
            If refCell IsNot Nothing Then
                newCell = refCell
            Else
                newCell = New Cell
                newCell.CellReference = cellReference
                row.InsertBefore(newCell, refCell)
            End If

            worksheet.Save()

            Return newCell
        End If
    End Function

    ''' <summary>
    ''' Delete text from a cell in an existing Spreadsheet Document
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="sheetName"></param>
    ''' <param name="colName"></param>
    ''' <param name="rowIndex"></param>
    ''' <remarks>
    ''' Given a document, a worksheet name, a column name, and a one-based row index,
    ''' deletes the text from the cell at the specified column and row on the specified sheet.
    ''' Column names are like "A", row index starts with 1
    ''' From: Open XML SDK 2.0
    ''' </remarks>
    Public Sub DeleteTextFromCell(ByVal document As SpreadsheetDocument, ByVal sheetName As String, ByVal colName As String, ByVal rowIndex As UInteger)
        Using (document)
            Dim sheets As IEnumerable(Of Sheet) = document.WorkbookPart.Workbook.GetFirstChild(Of Sheets)().Elements(Of Sheet)().Where(Function(s) s.Name = sheetName.ToString())
            If (sheets.Count = 0) Then
                ' The specified worksheet does not exist.
                Return
            End If

            Dim relationshipId As String = sheets.First.Id.Value
            Dim worksheetPart As WorksheetPart = CType(document.WorkbookPart.GetPartById(relationshipId), WorksheetPart)

            ' Get the cell at the specified column and row.
            Dim cell As Cell = GetSpreadsheetCell(worksheetPart.Worksheet, colName, rowIndex)
            If (cell Is Nothing) Then
                ' The specified cell does not exist.
                Return
            End If

            cell.Remove()
            worksheetPart.Worksheet.Save()

            ' Clean up the shared string table.
            If (cell.DataType.Value = CellValues.SharedString) Then
                Dim shareStringId As Integer = Integer.Parse(cell.CellValue.Text)
                RemoveSharedStringItem(shareStringId, document)
            End If

        End Using
    End Sub
    ''' <summary>
    ''' Get the cell at the column name and row index specified
    ''' </summary>
    ''' <param name="worksheet"></param>
    ''' <param name="columnName"></param>
    ''' <param name="rowIndex"></param>
    ''' <returns></returns>
    ''' <remarks>
    ''' Given a worksheet, a column name, and a row index, gets the cell at the specified column and row.
    ''' Column names are like "A", row index starts with 1
    ''' From: Open XML SDK 2.0
    ''' </remarks>
    Private Function GetSpreadsheetCell(ByVal worksheet As Worksheet, ByVal columnName As String, ByVal rowIndex As UInteger) As Cell
        Dim rows As IEnumerable(Of Row) = worksheet.GetFirstChild(Of SheetData)().Elements(Of Row)().Where(Function(r) r.RowIndex = rowIndex.ToString())
        If (rows.Count = 0) Then
            ' A cell does not exist at the specified row.
            Return Nothing
        End If

        Dim cells As IEnumerable(Of Cell) = rows.First().Elements(Of Cell)().Where(Function(c) String.Compare(c.CellReference.Value, columnName + rowIndex.ToString(), True) = 0)
        If (cells.Count = 0) Then
            ' A cell does not exist at the specified column, in the specified row.
            Return Nothing
        End If

        Return cells.First
    End Function
    ''' <summary>
    ''' Remove an unused shared string from a Spreadsheet Document
    ''' </summary>
    ''' <param name="shareStringId"></param>
    ''' <param name="document"></param>
    ''' <remarks>
    ''' Given a shared string ID and a SpreadsheetDocument, verifies that other cells in the document no longer
    ''' reference the specified SharedStringItem and removes the item.
    ''' From: Open XML SDK 2.0
    ''' </remarks>
    Private Sub RemoveSharedStringItem(ByVal shareStringId As Integer, ByVal document As SpreadsheetDocument)
        Dim remove As Boolean = True

        For Each part As WorksheetPart In document.WorkbookPart.GetPartsOfType(Of WorksheetPart)()
            Dim worksheet As Worksheet = part.Worksheet

            For Each cell As Cell In worksheet.GetFirstChild(Of SheetData)().Descendants(Of Cell)()
                ' Verify if other cells in the document reference the item.
                If (Not (cell.DataType) Is Nothing AndAlso cell.DataType.Value = CellValues.SharedString AndAlso cell.CellValue.Text = shareStringId.ToString()) Then
                    ' Other cells in the document still reference the item. Do not remove the item.
                    remove = False
                    Exit For
                End If
            Next

            If (Not remove) Then
                Exit For
            End If
        Next

        ' Other cells in the document do not reference the item. Remove the item.
        If (remove) Then
            Dim shareStringTablePart As SharedStringTablePart = document.WorkbookPart.SharedStringTablePart
            If (shareStringTablePart Is Nothing) Then
                Return
            End If

            Dim item As SharedStringItem = shareStringTablePart.SharedStringTable.Elements(Of SharedStringItem)().ElementAt(shareStringId)
            If (Not (item) Is Nothing) Then
                item.Remove()

                ' Refresh all the shared string references.
                For Each part As WorksheetPart In document.WorkbookPart.GetPartsOfType(Of WorksheetPart)()
                    Dim worksheet As Worksheet = part.Worksheet

                    For Each cell As Cell In worksheet.GetFirstChild(Of SheetData)().Descendants(Of Cell)()
                        If (Not (cell.DataType) Is Nothing AndAlso cell.DataType.Value = CellValues.SharedString) Then
                            Dim itemIndex As Integer = Integer.Parse(cell.CellValue.Text)
                            If (itemIndex > shareStringId) Then
                                cell.CellValue.Text = (itemIndex - 1).ToString()
                            End If
                        End If
                    Next

                    worksheet.Save()
                Next
                document.WorkbookPart.SharedStringTablePart.SharedStringTable.Save()
            End If
        End If
    End Sub

    ''' <summary>
    ''' Create or return a shared string with the specified text
    ''' </summary>
    ''' <param name="text"></param>
    ''' <param name="spreadSheetDocument">SpreadsheetDocument</param>
    ''' <returns>index of shared string</returns>
    ''' <remarks>
    ''' From: Open XML SDK 2.0 - modified to take SpreadSheetDocument as parameter
    ''' Given text and a SharedStringTablePart, creates a SharedStringItem with the specified text 
    ''' and inserts it into the SharedStringTablePart. If the item already exists, returns its index.
    ''' </remarks>
    Private Function InsertSharedStringItem(ByVal text As String, ByVal spreadSheetDocument As SpreadsheetDocument) As Integer
        Dim sharedStringPart As SharedStringTablePart

        If (spreadSheetDocument.WorkbookPart.GetPartsOfType(Of SharedStringTablePart).Count() > 0) Then
            sharedStringPart = spreadSheetDocument.WorkbookPart.GetPartsOfType(Of SharedStringTablePart).First()
        Else
            sharedStringPart = spreadSheetDocument.WorkbookPart.AddNewPart(Of SharedStringTablePart)()
        End If

        ' If the part does not contain a SharedStringTable, create one.
        If (sharedStringPart.SharedStringTable Is Nothing) Then
            sharedStringPart.SharedStringTable = New SharedStringTable
        End If

        Dim i As Integer = 0

        ' Iterate through all the items in the SharedStringTable. If the text already exists, return its index.
        For Each item As SharedStringItem In sharedStringPart.SharedStringTable.Elements(Of SharedStringItem)()
            If (item.InnerText = text) Then
                Return i
            End If
            i = (i + 1)
        Next

        ' The text does not exist in the part. Create the SharedStringItem and return its index.
        sharedStringPart.SharedStringTable.AppendChild(New SharedStringItem(New DocumentFormat.OpenXml.Spreadsheet.Text(text)))
        sharedStringPart.SharedStringTable.Save()

        Return i
    End Function
    ''' <summary>
    ''' Given a cell name, parses the specified cell to get the row index.
    ''' </summary>
    ''' <param name="cellName"></param>
    ''' <returns></returns>
    ''' <remarks>
    ''' From: Open XML SDK 2.0
    ''' Column names are like "A", row index starts with 1 
    ''' </remarks>
    Private Function GetRowIndex(ByVal cellName As String) As UInteger
        ' Create a regular expression to match the row index portion the cell name.
        Dim regex As Regex = New Regex("\d+")
        Dim match As Match = regex.Match(cellName)
        Return UInteger.Parse(match.Value)
    End Function
    ''' <summary>
    ''' Given a cell name, parses the specified cell to get the column name.
    ''' </summary>
    ''' <param name="cellName"></param>
    ''' <returns></returns>
    ''' <remarks>
    ''' From: Open XML SDK 2.0
    ''' Column names are like "A", row index starts with 1
    ''' </remarks>
    Private Function GetColumnName(ByVal cellName As String) As String
        ' Create a regular expression to match the column name portion of the cell name.
        Dim regex As Regex = New Regex("[A-Za-z]+")
        Dim match As Match = regex.Match(cellName)
        Return match.Value
    End Function
    '   Implicit Cell Formats (ECMA-376 Part 1, 18.8.30)
    '       0       General 
    '       1       0 
    '       2       0.00 
    '       3       #,##0 
    '       4       #,##0.00 
    '       9       0% 
    '       10       0.00% 
    '       11       0.00E+00 
    '       12       # ?/? 
    '       13       # ??/?? 
    '       14       mm-dd-yy 
    '       15       d-mmm-yy 
    '       16       d-mmm 
    '       17       mmm-yy 
    '       18       h:mm AM/PM 
    '       19       h:mm:ss AM/PM 
    '       20       h:mm 
    '       21       h:mm:ss 
    '       22       m/d/yy h:mm 
    '       37       #,##0 ;(#,##0) 
    '       38       #,##0 ;[Red](#,##0) 
    '       39       #,##0.00;(#,##0.00) 
    '       40       #,##0.00;[Red](#,##0.00) 
    '       45       mm:ss 
    '       46       [h]:mm:ss 
    '       47       mmss.0 
    '       48       ##0.0E+0 
    '       49       @ 
    Public Function GetImplicitCellFormat(ByVal formatNumber As Integer) As String
        If Formats.Count > formatNumber Then
            Return Formats(formatNumber)
        Else
            Return Formats(0)
        End If
    End Function
    ' Implicit Excel formats
    Dim Formats() As String = New String() {"General", "0", "0.00", "#,##0", "#,##0.00", "", "", "", "", "0%", "0.00%",
                                            "0.00E+00", "# ?/?", "# ??/??", "mm-dd-yy", "d-mmm-yy", "d-mmm", "mmm-yy", "h:mm AM/PM", "h:mm:ss AM/PM", "h:mm",
                                            "h:mm:ss", "m/d/yy h:mm", "", "", "", "", "", "", "", "",
                                           "", "", "", "", "", "", "#,##0 ;(#,##0)", "#,##0 ;[Red](#,##0)", "#,##0.00;(#,##0.00)", "#,##0.00;[Red](#,##0.00)",
                                            "", "", "", "", "mm:ss", "[h]:mm:ss", "mmss.0", "##0.0E+0", "@"}

    ' Special format conversions from Excel format strings to .NET
    Dim DateFormats As Dictionary(Of String, String) = New Dictionary(Of String, String) From {{"mm-dd-yy", "M/dd/yyyy"}, {"d-mmm-yy", "d-MMM-yy"}, {"m/d/yy h:mm", "M/d/yy h:mm"}}

    ''' <summary>
    ''' Convert an Excel format date number to a .NET value
    ''' </summary>
    ''' <param name="excelDate"></param>
    ''' <returns></returns>
    ''' <remarks>
    ''' Ignores special case in March 1900
    ''' </remarks>
    Public Function ConvertExcelDateToDateTime(ByVal excelDate As String) As DateTime
        Dim Days As Integer = CInt(excelDate)
        Dim Start As New DateTime(1900, 1, 1)
        Dim Result As DateTime = Start.AddDays(Days - 2)    'Excel thinks 1900 is a leap year
        Dim TimePart As Decimal = CDec(excelDate) - Days
        Dim HoursPart As Double = TimePart * 24
        Return Result.AddHours(HoursPart)
    End Function
    ''' <summary>
    ''' Convert a .NET DateTime to an excel serial number string
    ''' </summary>
    ''' <param name="dateValue"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function ConvertDateTimeToExcelDate(ByVal dateValue As DateTime) As String
        Dim Start As New DateTime(1900, 1, 1)
        Dim Days As Double = Convert.ToDouble(DateDiff(DateInterval.Day, Start, dateValue)) + 2
        Dim HoursPart As Double = dateValue.Hour / 24 + dateValue.Minute / (24 * 60)
        Return (Days + HoursPart).ToString
    End Function
    ''' <summary>
    ''' Close an existing document if open
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub CloseSpreadSheetDocument()
        If Me.SpreadSheetDocument Is Nothing Then Exit Sub
        Me.SpreadSheetDocument.Close()
        Me.SpreadSheetDocument.Dispose()
        Me.SpreadSheetDocument = Nothing
    End Sub
    ''' <summary>
    ''' Dispose the COM objects
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If Not Me.IsDisposed Then
            If disposing Then
                'free other state (managed objects)
            End If
            ' Clean up unmanaged objects
            GC.SuppressFinalize(Me)
            Me.IsDisposed = True
        End If
    End Sub

    ''' <summary>
    ''' Dispose if called
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub Finalize()
        Dispose(True)
    End Sub

End Class