Imports System.IO
Imports DocumentFormat.OpenXml.Packaging
Imports System.IO.Packaging
Imports System.Text
Public MustInherit Class OfficeInterface
    Implements IDisposable
    Public Sub New()

    End Sub
    ''' <summary>
    ''' Path to an Office document or template
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Document() As String

    ''' <summary>
    ''' Path to save an Office document
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks>
    ''' Using OpenXml it is no longer necessary to have the target application open
    ''' </remarks>
    Public Property SaveAs As String
 

    ''' <summary>
    ''' A dictionary of names and values to fill in
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property Fields() As Dictionary(Of String, String)
 
    Public Property DocumentPackage As Package

    Public Property PackagePart As PackagePart

    Public Property DocumentStream As MemoryStream

    Public Property IsDisposed() As Boolean
 
    Public MustOverride Sub Close()
    Public MustOverride Function FillInDocument() As Boolean
    Public MustOverride Function FillInDocument(ByVal document As String, ByVal fields As Dictionary(Of String, String)) As Boolean
    Public MustOverride Function GetFieldNames(ByVal document As String) As Dictionary(Of String, String)

    ''' <summary>
    ''' Load the starting file (template) into memory
    ''' </summary>
    ''' <param name="documentPath"></param>
    ''' <remarks></remarks>
    Public Sub LoadDocumentToMemoryStream(ByVal documentPath As String)
        If Document Is Nothing Then Throw New ArgumentNullException("documentPath")
        If Not File.Exists(documentPath) Then Throw New FileNotFoundException(documentPath)
        Dim Bytes() As Byte = File.ReadAllBytes(documentPath)
        DocumentStream = New MemoryStream()
        DocumentStream.Write(Bytes, 0, Bytes.Length)
    End Sub
    ''' <summary>
    ''' Close an existing document package if open
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub CloseDocumentPackage()
        If Me.DocumentPackage Is Nothing Then Exit Sub
        DocumentPackage.Close()
        Me.DocumentPackage = Nothing
    End Sub
    ''' <summary>
    ''' Concatenate title, first name, last name into Full Name
    ''' </summary>
    ''' <param name="fields"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetFullName(ByVal fields As Dictionary(Of String, String)) As String
        Dim sbldr As New StringBuilder
        If fields.ContainsKey("Title") Then sbldr.AppendFormat("{0} ", fields("Title"))
        If fields.ContainsKey("FirstName") Then sbldr.AppendFormat("{0} ", fields("FirstName"))
        If fields.ContainsKey("LastName") Then sbldr.Append(fields("LastName"))
        Return sbldr.ToString.Trim
    End Function
    ''' <summary>
    ''' Concatenate address components into a full address
    ''' </summary>
    ''' <param name="fields"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function GetFullAddress(ByVal fields As Dictionary(Of String, String)) As String
        Dim sbldr As New StringBuilder
        If fields.ContainsKey("Address1") Then sbldr.AppendFormat("{0}{1}", fields("Address1"), Environment.NewLine)
        If fields.ContainsKey("Address2") Then sbldr.AppendFormat("{0}{1}", fields("Address2"), Environment.NewLine)
        If fields.ContainsKey("City") Then sbldr.AppendFormat("{0}", fields("City"))
        If fields.ContainsKey("State") Then sbldr.AppendFormat("{0}{1}", ", ", fields("State"))
        If fields.ContainsKey("Zip") Then sbldr.AppendFormat(" {0}", fields("Zip"))
        Return sbldr.ToString.Trim
    End Function
    ''' <summary>
    ''' Writes the contents of a part into a file
    ''' </summary>
    ''' <param name="part">OpenXml part to read contents from</param>
    ''' <param name="filePath">Full path of file to write part contents to</param>
    Public Sub SavePartAs(ByVal part As OpenXmlPart, ByVal filePath As String)
        Dim partStream As Stream = part.GetStream(FileMode.Open, FileAccess.Read)
        Dim partContent As Byte() = New Byte(partStream.Length - 1) {}
        partStream.Read(partContent, 0, CInt(partStream.Length))
        File.WriteAllBytes(filePath, partContent)
    End Sub
    ''' <summary>
    ''' Save the current document to the provided path
    ''' </summary>
    ''' <param name="path"></param>
    ''' <remarks></remarks>
    Public Sub SaveDocument(ByVal path As String)
        If Document Is Nothing Then Throw New ArgumentNullException("Document")
        If String.IsNullOrEmpty(path) Then Throw New ArgumentNullException("Path")
        Me.SaveAs = path
        SaveDocument()
    End Sub
    ''' <summary>
    ''' Save the current document
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub SaveDocument()
        'Save the document if there is a "SaveAs" path
        If String.IsNullOrEmpty(Me.SaveAs) Then Exit Sub
        Dim Directory As String = Path.GetDirectoryName(Me.SaveAs)
        If Not IO.Directory.Exists(Directory) Then Throw New DirectoryNotFoundException(Directory)
        Using OutputStream As New FileStream(Me.SaveAs, FileMode.Create)
            DocumentStream.WriteTo(OutputStream)
            OutputStream.Close()
            DocumentStream.Close()
        End Using
    End Sub

    ' IDisposable
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.IsDisposed Then
            If disposing Then
                ' free other state (managed objects).
            End If

            ' free your own state (unmanaged objects).
            ' set large fields to null.
        End If
        Me.IsDisposed = True
    End Sub

#Region " IDisposable Support "
    ' This code added by Visual Basic to correctly implement the disposable pattern.
    Public Sub Dispose() Implements IDisposable.Dispose
        ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class

