﻿Imports System.IO
Imports System.IO.Packaging
Imports DocumentFormat.OpenXml.Packaging
Imports DocumentFormat.OpenXml.Wordprocessing
Imports DocumentFormat.OpenXml
''' <summary>
''' Provides the ability to fill in a Word document or template
''' </summary>
''' <remarks></remarks>
Public Class WordInterface
    Inherits OfficeInterface
    Private _WordDocument As WordprocessingDocument

    ''' <summary>
    ''' Word Document 
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property WordDocument() As WordprocessingDocument
        Get
            Return _WordDocument
        End Get
    End Property
    Private _Properties As Dictionary(Of String, String)
    ''' <summary>
    ''' Built in document properties collection
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public ReadOnly Property BuiltInDocumentProperties() As Dictionary(Of String, String)
        Get
            Return _Properties
        End Get
    End Property
    ''' <summary>
    ''' Default Constructor
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
        MyBase.new()
    End Sub
    ''' <summary>
    ''' Document only constructor
    ''' </summary>
    ''' <param name="document"></param>
    ''' <remarks></remarks>
    Public Sub New(ByVal document As String)
        MyBase.new()
        Me.Document = document
    End Sub
    ''' <summary>
    ''' Full Constructor
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="fields"></param>
    ''' <remarks></remarks>
    Public Sub New(ByVal document As String, ByVal fields As Dictionary(Of String, String))
        MyBase.new()
        Me.Document = document
        Me.Fields = fields
        ' Mark document as not to be autosaved - need to save a new version
        Dim Settings As New DocumentFormat.OpenXml.Packaging.OpenSettings
        Settings.AutoSave = False
        _WordDocument = CreateWordDocument(document, True, Settings)
    End Sub
    ''' <summary>
    ''' Close Document
    ''' </summary>
    ''' <remarks></remarks>
    Public Overrides Sub Close()
        CloseWordDocument()
        CloseDocumentPackage()
    End Sub
    ''' <summary>
    ''' Fill in a Word document with corresponding Fields values - document path and fields provided
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="fields"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function FillInDocument(ByVal document As String, ByVal fields As System.Collections.Generic.Dictionary(Of String, String)) As Boolean
        Me.Document = document
        Me.Fields = fields
        ' Mark document as not to be autosaved - need to save a new version
        Dim Settings As New DocumentFormat.OpenXml.Packaging.OpenSettings
        Settings.AutoSave = False
        _WordDocument = CreateWordDocument(document, True, Settings)
        Return FillInDocument()
    End Function
    ''' <summary>
    ''' Fill in the document
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function FillInDocument() As Boolean
        Dim Status As Boolean = False
        If Fields Is Nothing Then Throw New ArgumentNullException("Fields")
        If WordDocument Is Nothing Then Throw New ArgumentNullException("WordDocument")
        ' Assign a reference to the existing document body. 
        Dim DocumentBody As Body = WordDocument.MainDocumentPart.Document.Body
        GetBuiltInDocumentProperties()
        Dim BookMarks As Dictionary(Of String, String) = Me.GetFieldNames(DocumentBody)

        ' Determine dictionary variables to use - based on bookmarks in the document matching Fields entries
        If BookMarks.Count > 0 Then
            For Each item As KeyValuePair(Of String, String) In BookMarks
                Dim BookMarkName As String = item.Key
                If Me.Fields.ContainsKey(BookMarkName) Then
                    SetBookMarkValueByElement(DocumentBody, BookMarkName, Fields(BookMarkName))
                Else
                    ' Handle special case(s)
                    Select Case item.Key
                        Case "FullName"
                            SetBookMarkValueByElement(DocumentBody, BookMarkName, GetFullName(Fields))
                        Case "FullAddress"
                            SetBookMarkValueByElement(DocumentBody, BookMarkName, GetFullAddress(Fields))
                    End Select
                End If
            Next
            Status = True
        Else
            Status = False
        End If

        If String.IsNullOrEmpty(Me.SaveAs) Then Return Status
        Me.SaveDocument(WordDocument)
        Return Status
    End Function
    ''' <summary>
    ''' Get the document properties collection
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub GetBuiltInDocumentProperties()
        If WordDocument IsNot Nothing Then
            If Me._Properties Is Nothing Then Me._Properties = New Dictionary(Of String, String)
            Me._Properties.Clear()
            For Each Element As OpenXmlElement In WordDocument.ExtendedFilePropertiesPart.RootElement.Descendants
                If Me._Properties.ContainsKey(Element.LocalName) Then Continue For
                Me._Properties.Add(Element.LocalName, Element.InnerText)
            Next
        End If
    End Sub
    ''' <summary>
    ''' Create a WordProcessing document in a memory stream from a file
    ''' </summary>
    ''' <param name="document"></param>
    ''' <param name="isEditable"></param>
    ''' <param name="settings"></param>
    ''' <remarks>
    ''' Side effects - creates a Package and PackagePart as well as the memory stream
    ''' </remarks>
    Private Function CreateWordDocument(ByVal document As String, ByVal isEditable As Boolean, Optional ByVal settings As OpenSettings = Nothing) As WordprocessingDocument
        Me.CloseWordDocument()
        Me.CloseDocumentPackage()
        Me.Document = document
        LoadDocumentToMemoryStream(document)
        Me.DocumentPackage = Package.Open(DocumentStream, FileMode.Open, FileAccess.ReadWrite)
        PackagePart = DocumentPackage.GetPart(New Uri("/word/document.xml", UriKind.Relative))
        If settings Is Nothing Then
            Return WordprocessingDocument.Open(DocumentPackage)
        Else
            Return WordprocessingDocument.Open(DocumentPackage, settings)
        End If
    End Function
    ''' <summary>
    ''' Return a dictionary of Bookmark names, values in the provided document
    ''' </summary>
    ''' <param name="documentPath"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Overrides Function GetFieldNames(ByVal documentPath As String) As Dictionary(Of String, String)
        If Not File.Exists(documentPath) Then Throw New FileNotFoundException(documentPath)
        Me.Document = documentPath
        _WordDocument = CreateWordDocument(documentPath, False)
        Dim BookMarkNames As New Dictionary(Of String, String)
        ' Extract bookmarks from the document body. 
        Dim DocumentBody As Body = WordDocument.MainDocumentPart.Document.Body
        Return GetFieldNames(DocumentBody)
    End Function
    Private Overloads Function GetFieldNames(ByVal documentBody As Body) As Dictionary(Of String, String)
        Dim BookMarkNames As New Dictionary(Of String, String)
        BookMarkNames = GetBookMarksByElement(documentBody)
        Return BookMarkNames
    End Function
    Public Overloads Function GetFieldNames() As Dictionary(Of String, String)
        If WordDocument Is Nothing Then Throw New ArgumentNullException("Document not defined")
        Dim BookMarkNames As New Dictionary(Of String, String)
        ' Extract bookmarks from the document body. 
        Dim DocumentBody As Body = WordDocument.MainDocumentPart.Document.Body
        Return GetFieldNames(DocumentBody)
    End Function
    ''' <summary>
    ''' Get BookMarks from an OpenXml document by recursive search
    ''' </summary>
    ''' <param name="element"></param>
    ''' <returns>Dictionary(of String, String)</returns>
    ''' <remarks xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main">>
    ''' This is what a simple bookmark might look like:
    ''' <w:bookmarkStart w:name="Title" w:id="0" /> 
    ''' <w:r w:rsidR="00857EBE">
    ''' <w:t>Mr.</w:t> 
    ''' </w:r>
    ''' <w:bookmarkEnd w:id="0" /> 
    ''' </remarks>
    Private Function GetBookMarksByElement(ByVal element As OpenXmlElement) As Dictionary(Of String, String)
        Dim BookMarks As New Dictionary(Of String, String)
        For Each BookMark As BookmarkStart In element.Descendants(Of BookmarkStart)()
            Dim Name As String = BookMark.Name
            If String.IsNullOrEmpty(Name) OrElse Name.StartsWith("_") Then Continue For
            Dim Value As String = String.Empty
            ' Look for a value in the bookmark...
            Dim RunNode As Run = BookMark.NextSibling(Of Run)()
            If RunNode IsNot Nothing Then
                Dim TextNode As Text = RunNode.GetFirstChild(Of Text)()
                If TextNode IsNot Nothing Then Value = TextNode.Text
            End If
            BookMarks.Add(Name, Value)
        Next BookMark
        Return BookMarks
    End Function
    ''' <summary>
    ''' Set a bookmark value, using descendants of the given OpenXmlElement
    ''' </summary>
    ''' <param name="element"></param>
    ''' <param name="name"></param>
    ''' <param name="value"></param>
    ''' <returns>True if bookmark is found and set</returns>
    ''' <remarks>
    ''' Search for "simple" bookmark
    ''' </remarks>
    Private Function SetBookMarkValueByElement(ByVal element As OpenXmlElement, ByVal name As String, ByVal value As String) As Boolean
        For Each BookMark As BookmarkStart In element.Descendants(Of BookmarkStart)()
            If BookMark.Name = name And Not name.StartsWith("_") Then
                ' Look for a value in the bookmark...
                Dim RunNode As Run = BookMark.NextSibling(Of Run)()

                ' Handle the special case of newline characters in the value string
                Dim Values As String() = value.Split(Environment.NewLine)

                If RunNode IsNot Nothing Then
                    Dim TextNode As Text = RunNode.GetFirstChild(Of Text)()
                    If TextNode IsNot Nothing Then
                        TextNode.Text = Values(0)
                    Else
                        TextNode = New Text(Values(0))
                        RunNode.AppendChild(Of Text)(TextNode)
                    End If
                Else
                    Dim TextNode As New Text(Values(0))
                    RunNode = New Run()
                    BookMark.InsertAfterSelf(RunNode)
                    RunNode.AppendChild(TextNode)
                End If
                If Values.Length > 1 Then
                    AppendMultiple(RunNode, Values)
                End If
                Return True
            End If
        Next BookMark

        Return False
    End Function
    ''' <summary>
    ''' Add additional values to the Run node
    ''' </summary>
    ''' <param name="runNode"></param>
    ''' <param name="values"></param>
    ''' <remarks>
    ''' If there were new line characters in the input, add breaks for each followed by the next value
    ''' </remarks>
    Private Sub AppendMultiple(ByVal runNode As Run, ByVal values As String())
        Dim i As Integer
        For i = 1 To values.Length - 1
            runNode.AppendChild(Of Break)(New Break)
            runNode.AppendChild(Of Text)(New Text(values(i)))
        Next
    End Sub
 ''' <summary>
    ''' Dispose
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If Not Me.IsDisposed Then
            If disposing AndAlso Me.WordDocument IsNot Nothing Then
                CloseWordDocument()
            End If
            GC.SuppressFinalize(Me)
            Me.IsDisposed = True
        End If
    End Sub
    ''' <summary>
    ''' Close an existing document if open
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub CloseWordDocument()
        If Me.WordDocument Is Nothing Then Exit Sub
        Try
            Me.WordDocument.Close()
            Me.WordDocument.Dispose()
        Catch ex As ObjectDisposedException
        Catch ex As System.ArgumentException
        Finally
            Me._WordDocument = Nothing
        End Try
    End Sub
    ''' <summary>
    ''' Save the current document to a file
    ''' </summary>
    ''' <remarks></remarks>
    Public Overloads Sub SaveDocument(ByVal document As WordprocessingDocument)
        'Save the document in memory before calling SaveDocument
        document.MainDocumentPart.Document.Save()
        MyBase.SaveDocument()
    End Sub
    ''' <summary>
    ''' Dispose if called
    ''' </summary>
    ''' <remarks></remarks>
    Protected Overrides Sub Finalize()
        Dispose(True)
    End Sub
End Class
